/*
    Copyright (C) 1999-2001 Rui Miguel Silva Seabra <rms@greymalkin.yi.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <syslog.h>

#include <getopt.h>

#define PROGRAM "irmctl"
#define VERSION "0.4.0"
#define LICENSE "GNU GPL 2.0"

#include "irmctl_readconf.h"
#include "libirmctl.h"

pIRMCTL conf=NULL;
gchar *configuration=NULL;
gchar *device=NULL;
IRMCTL_CODE * code=NULL;

void sigint_handle(gint s);
gint readconf(gchar* path);


void sigusr1_handle(gint s) {
	signal(SIGUSR1, SIG_IGN);
	syslog(LOG_INFO, "Rereading Configuration");
	if(!readconf(configuration)) {
		sigint_handle(EXIT_FAILURE);
	}
	signal(SIGUSR1, sigusr1_handle);
}
void sigint_handle(gint s) {
	irmctl_finish_device(device);
	syslog(LOG_INFO, "Stopping services");
	closelog();
	if(s==EXIT_FAILURE) exit(s);
	else exit(0);
}
void sigcld_handle(gint s) {
	gint status;
	wait(&status);
	if(signal(SIGCLD, sigcld_handle) == SIG_ERR)
#ifdef DEBUG
		g_warning("Handler for SIGCLD not set!");
#else
		syslog(LOG_WARNING, "Handler for SIGCLD not set!");
#endif
}

void irmctl_switch_mode() {
	pMODE mode=NULL;
#ifdef DEBUG
	g_message("Switch pressed.\n");
#endif
	// read mode switch
	code=irmctl_get_code();
	if(code) {
		mode=g_hash_table_lookup(conf->modes,
				g_hash_table_lookup(conf->tc->nameof,
					code));
		if(mode) {
			conf->currentMode=mode;
#ifdef DEBUG
			g_message("Mode changed: %s.\n", mode->name);
#endif
		}
		else {
#ifdef DEBUG
			g_warning("Mode does not exist.\n");
#else
			syslog(LOG_WARNING, "Mode does not exist.\n");
#endif
		}
	}
}

void irmctl_exec(gchar ** tokens) {
	pid_t pid;
	struct stat buf;

#ifdef DEBUG
	g_message("Command: %s\n", (gchar*)g_strjoinv(" ", tokens));
#endif
	if(-1 == stat(tokens[0], &buf)) {
		syslog(LOG_WARNING, "Refusing command: %s", sys_errlist[errno]);		return;
	}

	if( (pid=fork())<0 )
#ifdef DEBUG
		g_warning("Unnable to fork()");
#else
		syslog(LOG_ERR, "Unnable to fork()");
#endif
	else if (!pid) {
		pid=execv(tokens[0], tokens);
#ifdef DEBUG
		g_warning("Couldn't exec: %s\n", sys_errlist[errno]);
#else
		syslog(LOG_WARNING, "Couldn't exec: %s", sys_errlist[errno]);
#endif
		exit(1);
	}
}

void irmctl_activate(pACTION action) {
	pFUNCTION func;
	if(action->type) {
		func=action->handler.func;
		switch(func->type) {
			case 0:
				func->args();
				break;
			case 1:
				func->args(getarg1(func));
				break;
			case 2:
				func->args(getarg1(func),getarg2(func));
				break;
			default:
				break;
		}
	} else irmctl_exec(action->handler.command);
}


void irmctl_handle_action(gpointer button) {
	pACTION action;
	GHashTable *actions=conf->currentMode?conf->currentMode->actions:NULL;

#ifdef DEBUG
	g_message("Exec'ing: %s\n", (gchar*)button);
#endif

	if(button && actions && (action=g_hash_table_lookup(actions, button)))
		irmctl_activate(action);
}


void irmctl_ten_mode(gint tenMult) {
	gchar *button, *endptr=NULL,*selection;
	glong but;

#define BUTTONLENGTH 50
	
#ifdef DEBUG
	g_warning("%d times Ten\n", tenMult);
#endif
	code=irmctl_get_code();
	if(!strcmp(code, conf->tenbutton)) {
		irmctl_code_free(code);
		if (tenMult < 9)
			irmctl_ten_mode(tenMult+1);
#ifdef DEBUG
		else
			g_message("oops... exceeded count of ten multiplier\n");
#endif
		return;
	}
	button=g_hash_table_lookup(conf->tc->nameof, code);
	if(!button)
		button="";
	but=strtol(button, &endptr, 10);
	if(!strcmp(endptr,button) || but < 0 || but > 9) {
#ifdef DEBUG
		g_message("oops... lost count of ten multiplier\n");
#endif
		irmctl_code_free(code);
		return;
	}
	but+=tenMult*10;
	selection=(gchar*)g_malloc(BUTTONLENGTH*sizeof(gchar));
	g_snprintf(selection, BUTTONLENGTH, "%d", (gint)but);
#ifdef DEBUG
	g_warning("should check for truncation in irmctl_ten_mode\n");
	g_message("You chose: %s\n", selection);
#endif
	irmctl_code_free(code);
	irmctl_handle_action(selection);
	g_free(selection);
}

void irmctl_handle(IRMCTL_CODE * code) {
	gchar *button;
#ifdef DEBUG
	gchar **aux;
	pACTION a;
#endif

	button=g_hash_table_lookup(conf->tc->nameof, code);
	// FIXME: ugly hack so it won't core dump when button not applyable
	if(!button) return;
	if(!strcmp(button, conf->switchbutton)) {
		irmctl_code_free(code);
		irmctl_switch_mode();
		return;
	}
	if(!strcmp(button, conf->tenbutton)) {
		irmctl_code_free(code);
		irmctl_ten_mode(1);
		return;
	}
#ifdef DEBUG
	//g_message("Pressed '%s' <=> '%s' button\n", code, button);
	if(conf->currentMode) {
		a=g_hash_table_lookup(conf->currentMode->actions, button);
		if(a) {
			if(a->type)
				g_message("Pressed '%s' <=> '%s'\n",
					  code, button);
			else {
				aux=a->handler.command;
				g_message("Pressed '%s' <=> '%s ==> %s'\n",
					  code, button, g_strjoinv(" ", aux));
			}
		}
		else g_warning("Button %s not defined\n", button);
	}
#endif
	irmctl_handle_action(button);
}

void irmctl_main(gchar* device_to_use) {
	gchar*device=device_to_use?device_to_use:"/dev/ttyS0";
#ifdef DEBUG
	gint i=1;
#endif

	if(signal(SIGCLD, sigcld_handle) == SIG_ERR)
		g_warning("Handler for SIGCLD not set!");
	if(signal(SIGINT, sigint_handle) == SIG_ERR)
		g_warning("Handler for SIGINT not set!");
	if(signal(SIGUSR1, sigusr1_handle) == SIG_ERR)
		g_warning("Handler for SIGUSR1 not set!");

	openlog(PROGRAM, LOG_PID, LOG_DAEMON);
	syslog(LOG_INFO, "Starting services");

	if(!irmctl_init_device(device)) {
#ifdef DEBUG
		g_warning("Device not ready: %s", device);
		g_error("Stopping services");
#else
		syslog(LOG_CRIT, "Device not ready: %s", device);
		syslog(LOG_INFO, "Stopping services");
		exit(EXIT_FAILURE);
#endif
	}
#ifdef DEBUG
	g_message("Using device: %s", device);
#else
	close(STDOUT_FILENO);
#endif

	/* Main event loop.
	 * can only exit by a signal or by a quit code.
	 * 
	 */
	while(1) {
#ifdef DEBUG
		g_message("Run #%d\n",i++);
#endif
		code=irmctl_get_code();
		if(code) irmctl_handle(code);
		else
#ifdef DEBUG
			g_warning("Not enough memory to read code!");
#else
			syslog(LOG_WARNING, "Not enough memory to read code!");
#endif
		irmctl_code_free(code);
			
	}
}

gint daemonize() {
	pid_t pid;
	if ((pid=fork()) < 0)
		return(-1);
	else if (pid!=0)
		exit(0);

	setsid();
	chdir("/");
	umask(0);
	return(0);
}

void irmctl_version() {
	g_print("%s %s, released under %s\n", PROGRAM, VERSION, LICENSE);
}
void irmctl_usage() {
	g_print("Usage: %s [options]\n",PROGRAM);
	g_print("\t -h --help\t\tdisplay this message\n");
	g_print("\t -h --help\t\tdisplay this message\n");
	g_print("\t --conf=/path/to/conf\n");
	g_print("\t --device=/path/to/device\n");
}

gint readconf(gchar* path) {
	if (path) {
		g_print("Reading configuration from '%s'\n", path);
		conf = irmctl_loadconf(path);
	} else {
		// FIXME: Read configuration, search /etc
		// and /usr/local/etc else fail?
		conf = irmctl_loadconf("/etc/irmctl.conf");
		g_message("Configuration file is '/etc/irmctl.conf'\n");
	}
	return(1);
}

void handle_option(gint option, gchar * argument) {
	size_t	len;

	switch(option) {
		case 'h':
			irmctl_usage();
			exit(EXIT_SUCCESS);
		case 'v':
			irmctl_version();
			exit(EXIT_SUCCESS);
		case 'd':
			len=((len=strlen(argument))>PATH_MAX?PATH_MAX:len)+1;
			device=g_strndup(argument, len);
			break;
		case 'c':
			len=((len=strlen(argument))>PATH_MAX?PATH_MAX:len)+1;
			configuration=g_strndup(argument, len);
			break;
		default:
			exit(EXIT_FAILURE);
	}
}

gint main(gint argc, gchar**argv) {
	gint c;
	static struct option long_options[] = {
		{"help", no_argument, NULL, 'h'},
		{"version", no_argument, NULL, 'v'},
		{"device", required_argument, NULL, 'd'},
		{"conf", required_argument, NULL, 'c'},
		{0, 0, 0, 0}
	};

	while(1) {
		c = getopt_long(argc, argv, "hvd:c:", long_options, NULL);
		if (c==-1)
			break;
		handle_option(c, optarg);
	}
	if (readconf(configuration)) {
#ifndef DEBUG
		daemonize();
#else
		g_message("Normally, daemonize() here");
#endif
		irmctl_main(device);
	}
	return(EXIT_FAILURE);
}
